/* DSmeister software
 * MCore.ino
 *
 * Created: 2015/09/06
 *  Author: Yaasan
 */ 

#include <avr/wdt.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_MM2.h"
#include "DSCoreM.h"
#include "DS_Buzzer.h"

/***********************************************************************
	
	モジュール内定数宣言
	
**********************************************************************/


#define		SIZE_PACKET 		19
#define		PACKET_NOTRECEIVED	0
#define		PACKET_RECEIVED		1
#define		PACKET_SENT			0
#define		PACKET_NOTSEND		1

#define		CMD_GET				0xA0
#define		CMD_SET				0xE0

#define		TIME_LPF_OC			128
#define		TIME_LPF_OV			100


/***********************************************************************
	
	モジュール内構造体宣言
	
**********************************************************************/

typedef union
{
	UC BYTE;
	struct {
		UC OC:1;
		UC LV:1;
		UC OV:1;
		UC OH:1;
		UC reserved:4;
	} BIT;
} FLAG_DSCORE;


/***********************************************************************
	
	モジュール内グローバル変数宣言
	
**********************************************************************/


UC gPacketMode; /**< モード */

/*
  パケット仕様：
  
0: IDENTIFICATION & SIZE(Max 3+16byte)
1: STATUS (ERROR)
2: COMMAND
3: DATA1
4: DATA2
5: DATA3
.
.
X: CHECKSUM

  
  
*/
UC gReceivePacket[SIZE_PACKET];
UC gSendPacket[SIZE_PACKET];
UC gPacketCount = 0;
UC gPacketReceived = PACKET_NOTRECEIVED;
UC gPacketSent = PACKET_SENT;

UC gPoweronTrigger = TRIGGER_NONE;
unsigned long gTime_100ms = 0;
FLAG_DSCORE gFlag_Error;
long gBufLPF_OC = 0;
long gBufLPF_OV = 0;


DSCoreLib DSCore;
DSBuzzer Buzzer;

/***********************************************************************
	
	モジュール内関数宣言
	
**********************************************************************/

void init();
void SPI_Init();

void Main_sendReplyPacket(UC inReplyCommand);
UC Main_decodeMain(UC *inPackets);

bool clearSafety();
bool checkSafety();

void CalculateOverCurrent();
bool GetOCValue();

/***********************************************************************
	
	関数実現部
	
**********************************************************************/

UC Main_decodeMain(UC *inPackets)
{
	UC aInLen;
	US aDecAddress;
	US aAddress;
	short aSpeed;
	UC aFunction;
	UC aPower;
	UC aDirection;
	UC aCRC;
	UC aCVValue;
	UC aProtocol = 0;
	UC aPowerOnEnd;
	UC aRet = 0x00;
	boolean aCVResult;
	
	
	/* サイズの取得 */
	aInLen = inPackets[0] & 0b1111;
	
	/* サイズチェック */
	if( aInLen <= 1)
	{
		/* CMD ERR */
		return 0xFE;
	}
	
	/* パケットチェック */
	aCRC = DS_CalculateCRC(inPackets, aInLen);
	
	if( aCRC != inPackets[aInLen - 1])
	{
		/* CRCエラー*/		
		//return 0xFF;
	}	

	/* コマンドごとに処理 */
	
	switch(inPackets[0] & 0b11110000)
	{
	case CMD_PWR_OFF:
		{
			
		DSCore.SetPower(0);
		gPoweronTrigger = TRIGGER_NONE;		
		
		//DEBUG
		Serial.println("[CMD] Power Off");
		break;
		}
	
	case CMD_PWR_ON:
		/* 線路に電源投入 */
		DSCore.SetPower(1);
		
		//DEBUG
		Serial.println("[CMD] Power On");
		
		break;	
			
	case CMD_SPEED:
		
		/* データ取得 */
		aAddress = (inPackets[2] << 8) + inPackets[1];
		aSpeed = (inPackets[4] << 8) + inPackets[3];
		
		if( aInLen >= 7)
		{
			/* SpeedStepのセット */
			aProtocol = inPackets[5];
		}
		else
		{
			/* デフォルト（DCC 28, MM2 14）の使用(書き換えない) */
			aProtocol = 0xFF;
		}
		
		/* アドレスデコード処理 */
		DSCore.SetLocoSpeedEx( aAddress, aSpeed, aProtocol);
		
		//DEBUG
		Serial.print("[CMD] ADR=");
		Serial.print(aAddress);
		Serial.print(",SPD=");
		Serial.print(aSpeed);
		Serial.print(",PRT=");
		Serial.println(aProtocol);
		
		break;
		
	case CMD_ACCESSORY:
		aAddress = (inPackets[2] << 8) + inPackets[1];
		//inPackets[3];は未使用。（Positionの予約だったが、使わないことになってしまった。)
		aPower = inPackets[4];
		
		/* アクセサリ指令 */
		DSCore.SetTurnout(aAddress, aPower);
		break;
		
	case CMD_FUNCTION:
		aAddress = (inPackets[2] << 8) + inPackets[1];
		aFunction = inPackets[3];
		aPower = inPackets[4];
		
		/* ファンクション指令 */
		DSCore.SetLocoFunction( aAddress, aFunction, aPower);
		
		break;
		
	case CMD_DIRECTION:
		/* データ取得 */
		aAddress = (inPackets[2] << 8) + inPackets[1];
		aDirection =  inPackets[3];// FWD:0 REV:1
		
	
		/* 進行方向指令 */
		DSCore.SetLocoDirection( aAddress, aDirection);
		
		break;
		
	case CMD_CVWRITE:
		/* CV書き込み */
		aAddress = ((inPackets[2] << 8) + inPackets[1]);
		aCVValue = inPackets[3];
		
		if( aInLen >= 7)
		{
			//Ops Mode用
			aDecAddress = ((inPackets[5] << 8) + inPackets[4]);
			DSCore.WriteConfig_Ops(aDecAddress, aAddress & 0b11111111111, aCVValue);
		}
		else
		{
			//Direct Mode用
			aDecAddress = 0;
			DSCore.WriteConfig_Dir(aAddress & 0b11111111111, aCVValue);
		}
		
		break;	
		
	case CMD_CVREAD:
		/* CV読み込み, Direct Mode専用(高速,回路がシンプルなため) */
		aAddress = ((inPackets[2] << 8) + inPackets[1]);
		aCVValue = 0;
		aPowerOnEnd = DS_IsPowerOn();
		
		//DEBUG
		Serial.println("[CMD] CV Read");

		aCVResult = DSCore.ReadConfig(0, aAddress & 0b1111111111, &aCVValue);
		
		gSendPacket[0] = CMD_CVREAD | 0b0010;
		gSendPacket[1] = aCVValue;
		gSendPacket[2] = (aCVResult == true) ? 1 : 0;
		gSendPacket[3] = 0x00;
	
		gPacketSent = PACKET_NOTSEND;
		gPacketCount = 3;
		aRet = CMD_CVREAD;
		
		break;

	case CMD_OK:
		//DEBUG
		Serial.println("[CMD] Receive OK");
		break;
	case CMD_WAIT:
		/* ウェイト */	
		//DEBUG
		Serial.println("[CMD] Receive WAIT");
	
		break;		
					
	default:
		/*コマンドエラーを返す*/
		return 0xFE;
		break;
	
	}
	
	return aRet;
		
}




void Main_sendReplyPacket(UC inReplyCommand)
{	 
	 
	 gSendPacket[0] = inReplyCommand;
	 gPacketCount = 0;
	 
	 /* 送信パケット初期化 */
	 for( int i = 1; i < 7; i++)
	 {
		 gSendPacket[i] = 0x00;
	 }
	 
	gPacketSent = PACKET_NOTSEND;
}


bool clearSafety()
{
	UC aFlag = gFlag_Error.BYTE;
	
	//Clear
	gFlag_Error.BYTE = 0;
	
	//Check
	checkSafety();
	
	if( gFlag_Error.BYTE != 0)
	{
		/* エラーランプ点灯 */
		DS_LEDERR(1);
		
		return false;
	}
	else
	{
		/* エラーランプ消灯 */
		DS_LEDERR(0);
		
		return true;
	}
	
}


bool checkSafety()
{
	bool aError = false;
	
	
	US aVoltage = gBufLPF_OV >> 8;
		
	if( aVoltage < LV_LEVEL)
	{
		//Low Voltage 9.0V or less
		gFlag_Error.BIT.LV = 1;
		
		Serial.print("Voltage=");
		Serial.print(aVoltage);
		Serial.println("");
		
		aError = true;
	}
		
	if( aVoltage >= OV_LEVEL)
	{
		//Over Voltage 25.5V or more
		gFlag_Error.BIT.OV = 1;
		
		Serial.print("Voltage=");
		Serial.print(aVoltage);
		Serial.println("");
		
		aError = true;
	}
	
	/* OCの取得 */
	US aCurrent = GetOCValue();
	
	if( aCurrent >= OC_LEVEL)
	{
		//Over Current 10.0A or more
		gFlag_Error.BIT.OC = 1;
		
		Serial.print("Current=");
		Serial.print(aVoltage);
		Serial.println("");

		aError = true;
	}
	
	return aError;

}

void loop()
{
	
	while(1)
	{
		
		/* Reset watchdog timer. */
		wdt_reset(); 
		
		
		//運転中のみ実行
		if( DS_IsPowerOn() == 1)
		{
			//保護機能動作
			if( checkSafety()== true)
			{
				
				/* エラーランプ点灯 */
				DS_LEDERR(1);
				
				//停止、エラー処理
				Serial.print("Error OC:");
				Serial.print(gFlag_Error.BIT.OC);
				Serial.print(" OV:");
				Serial.print(gFlag_Error.BIT.OV);
				Serial.print(" LV:");
				Serial.print(gFlag_Error.BIT.LV);
				Serial.print(" OH:");
				Serial.print(gFlag_Error.BIT.OH);
				Serial.println("");
				
				//Power off
				DSCore.SetPower(0);
				gPoweronTrigger = TRIGGER_NONE;
				
				//DEBUG
				Serial.println("[CMD] Power Off");
				
				/* 起動ブザーを鳴らす */
				Buzzer.Register(2, 36);
				
			}
		}
		
		/* 受信確認 */
		if( gPacketReceived == PACKET_RECEIVED)
		{
			/* データデコード&リスト更新 or 登録 */
			UC aSPIError = Main_decodeMain(gReceivePacket);
			
			/* 送信データ生成（応答処理） */
			switch( aSPIError)
			{
				/*処理完了*/
				case 0x00: 
					Main_sendReplyPacket(CMD_OK);
					break;
				/* 指令を待っている状態 */
				case 0x01:
					Main_sendReplyPacket(CMD_WAIT);
					break;
				/* コマンドが一致しなかった */			
				case 0xFE:
					Main_sendReplyPacket(CMD_CMDERR);
					/* エラーランプ点灯 */
					DS_LEDERR(1);
					break;
				/* CRCが一致しなかった */
				case 0xFF:
					Main_sendReplyPacket(CMD_CRCERR);
					/* エラーランプ点灯 */
					DS_LEDERR(1);					
					break;
				/* 原因不明のエラー */
				case CMD_CVREAD:
					/* CV値応答 */
					
					break;
				default:
					Main_sendReplyPacket(CMD_UNKERR);
					/* エラーランプ点灯 */
					DS_LEDERR(1);
					break;				
			}
			
			/* パケット受信フラグのリセット */
			gPacketReceived = PACKET_NOTRECEIVED;
		}
		
		//パルスジェネレータのスキャン
		DSCore.Scan();
		
		//ブザーのスキャン
		Buzzer.Scan();
		
		// 100ms interval
		if( (millis() - gTime_100ms) >= 100)
		{
			//OC Calculation
			CalculateOverCurrent();
			
			
			
			//Reset task
			gTime_100ms = millis();
			
			
		}
	}
}




void CalculateOverCurrent()
{
	
	long aVoltage = analogRead(SENSOR_VOLTAGE);
	//	Serial.println(aVoltage);
	
	//LPF
	gBufLPF_OV = ((gBufLPF_OV * (256 - TIME_LPF_OV)) >> 8) + ( TIME_LPF_OV * aVoltage);
	
	
	//この関数は100us程度かかる
	long aCurrent = analogRead(SENSOR_CURRENT);
	
	/* 電流値の計算(512dを基準に絶対値を取る) */
	
	if( aCurrent >= 512)
	{
		aCurrent = aCurrent - 512;
	}
	else
	{
		aCurrent = 512 - aCurrent;
	}
	
	
	gBufLPF_OC = ((gBufLPF_OC * (256 - TIME_LPF_OC)) >> 8) + ( TIME_LPF_OC * aCurrent);

}

bool GetOCValue()
{
	return gBufLPF_OC >> 8;
}


void setup()
{
	DSCore.Init();
	Buzzer.Init();
	
	
	Serial.begin(115200);
	while (!Serial);

	Serial.println("Desktop Station Meister Core R0");
	
	SPI_Init();
	
	/* LED制御 */
	DDRC =  (1<<DDC0)|(1<<DDC1);
	PORTC = 0b00000011;
	
	/* ゲート信号制御 */
	DDRD =  (1<<DDD6)|(1<<DDD5);
	PORTD = 0b00011100;	/* INT0, INT1, PCINT20のプルアップをON */
	
	/* 外部INT割込み */
	EICRA = 0b0010;
	EIMSK = 0b01;
	
	/* PCINT0(SPI SSの論理変化割込み) */
	PCMSK0 = 0b00000100;
	PCICR = 0b101;
	
	/* PCINT20(RESTARTの論理変化割込み) */
	PCMSK2 = 0b00010000;
	
	
	/* TIMER1設定 */
	TCCR1A = (0<<COM1A1) | (1<<COM1A0) | (0<<COM1B1) | (1<<COM1B0) | (0<<WGM11) | (0<<WGM10);
	TCCR1B = (0<<ICNC1)  | (0<<ICES1)  | (0<<WGM13)  | (1<<WGM12)  | (0<<CS12)  | (1<<CS11) | (0<<CS10);
	OCR1A = OCR1B = 115; /* 58us */
	TCNT1 = 0;
	
	/* 保護センサ系 */
	gFlag_Error.BIT.OC =0;
	gFlag_Error.BIT.OV =0;
	gFlag_Error.BIT.LV =0;
	gFlag_Error.BIT.OH =0;
	

	/* LED_RUN ON */
	DS_LEDRUN(0);
			
	/* エラーランプ解除 */
	DS_LEDERR(0);
	
	/* ウォッチドッグ有効(8sec) */
	wdt_enable(WDTO_8S); // WDT reset
	
	/* 割込み許可 */
	sei();
	
	
	/* 応答パケットを記載 */
	Main_sendReplyPacket(CMD_WAIT);
	
	
	/* 起動ブザーを鳴らす */
	Buzzer.Register(1, 12);
	
	//Reset task
	gTime_100ms = millis();	
	
	


	
}




void SPI_Init()
{

	// MISOを出力に
	DDR_SPI |= (1<<4);
	//MOSI,SCK,SSを入力に
	DDR_SPI &= 0xff - (1<<3) - (1<<2) - (1<<5);

	SPCR = 0b11100011;//SPI mode0(CPOL=0,CPHA=0,SLAVE,SPIenable,SPIIRWenable,,
	SPSR = 0b00000000;//倍速無効

}


ISR(SPI_STC_vect){			/* データが届いたとき＝SPIFが立ったときの割込みルーチン */

	if( gPacketCount < 7)
	{
		//次の通信で送るデータをセット
		SPDR = gSendPacket[gPacketCount];
	}
	
	//今回の通信で受信するデータ
	gReceivePacket[gPacketCount] = SPDR;	//１バイト取り込み

	gPacketCount++;


}

/* Emergencyボタンが押されたときの処理（強制パワーオフ） */
ISR(INT0_vect)
{
	/* チャタリング防止処理 */
	_delay_ms(10);
	
	/* 信号値をチェック */
	if( bit_is_clear(PIND, PD2) )
	{		
		
		/* 緊急停止とする */
		DSCore.SetPower(0);
		
		/* Arduino側にはオフのことを送信 */
		Main_sendReplyPacket(CMD_PWR_OFF);
	
	}
	else
	{
		/* 離したとき */
		
	}
	
}

ISR(PCINT0_vect)
{
	/* SPI SS信号割込み */
	
	if( bit_is_clear(PINB, PB2))
	{
		/* SSが立下り(通信開始) */
		gPacketCount = 0;
		
		/* パケット受信フラグ初期化 */
		gPacketReceived = PACKET_NOTRECEIVED;
	}
	else
	{
		/* SSが立上り（通信終了） */
		
		/* きちんと受信し終わっていたら */
		if(gPacketCount > 0)
		{
			/* パケット受信フラグのセット */
			gPacketReceived = PACKET_RECEIVED;			
		}
	
		
	}	
}

ISR(PCINT2_vect)
{
	/* チャタリング防止処理 */
	_delay_ms(10);
	


	
}

/*

ISR(TIMER1_COMPA_vect)
{
	
	
	
}




void DCCGEN_Start()
{
	TIMSK1 |= (1<<OCIE1A);
}

	
void DCCGEN_Start()
{
	TIMSK1 |= (1<<OCIE1A);
}
*/
	